/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <math.h>
#include <assert.h>
#include <limits.h>

#include "utils.h"
#include "value.h"
#include "variable.h"
#include "interpreter.h"
#include "builtin.h"
#include "builtin_gcode.h"
#include "builtin_svg.h"
#include "builtin_dxf.h"

#define WARNSHIFTS	1000	/* Trying to shift this many times may be an error */
#define NSORTFUNC	16	/* Fewer than about 16 functions is faster using linear search */

static value_t *interpret(node_t *n, node_t **brkref);
static void interpret_expr(node_t *n);

typedef value_t *(*opfunc_t)(const node_t *ref, value_t *lv, const value_t *rv);

static node_t **functions;
static int nfunctions;
static int nafunctions;

static const char *nodetypename(node_et n)
{
	switch(n) {
	case NT_INVALID:	return "NT_INVALID";
	case NT_EXPR:		return "NT_EXPR";
	case NT_EXPRLIST:	return "NT_EXPRLIST";
	case NT_IF:		return "NT_IF";
	case NT_ELIF:		return "NT_ELIF";
	case NT_FOR:		return "NT_FOR";
	case NT_FOREACH:	return "NT_FOREACH";
	case NT_REPEAT:		return "NT_REPEAT";
	case NT_WHILE:		return "NT_WHILE";
	case NT_DOWHILE:	return "NT_DOWHILE";
	case NT_FUNCTION:	return "NT_FUNCTION";
	case NT_ARGLIST:	return "NT_ARGLIST";
	case NT_LOCALLIST:	return "NT_LOCALLIST";
	case NT_CALL:		return "NT_CALL";
	case NT_BREAK:		return "NT_BREAK";
	case NT_CONTINUE:	return "NT_CONTINUE";
	case NT_RETURN:		return "NT_RETURN";
	case NT_LOCAL:		return "NT_LOCAL";
	}
	return "<unknown node type>";
}

static const char *expropname(op_et n)
{
	switch(n) {
	case OP_NULL:		return "OP_NULL";
	case OP_ADD:		return "OP_ADD";
	case OP_ADDOR:		return "OP_ADDOR";
	case OP_SUB:		return "OP_SUB";
	case OP_SUBOR:		return "OP_SUBOR";
	case OP_MUL:		return "OP_MUL";
	case OP_DIV:		return "OP_DIV";
	case OP_MOD:		return "OP_MOD";
	case OP_INDEX:		return "OP_INDEX";
	case OP_LOR:		return "OP_LOR";
	case OP_LAND:		return "OP_LAND";
	case OP_EQ:		return "OP_EQ";
	case OP_NE:		return "OP_NE";
	case OP_GT:		return "OP_GT";
	case OP_LT:		return "OP_LT";
	case OP_GE:		return "OP_GE";
	case OP_LE:		return "OP_LE";
	case OP_SHL:		return "OP_SHL";
	case OP_SHR:		return "OP_SHR";
	case OP_NOT:		return "OP_NOT";
	case OP_INT:		return "OP_INT";
	case OP_FLOAT:		return "OP_FLOAT";
	case OP_STRING:		return "OP_STRING";
	case OP_VECTOR:		return "OP_VECTOR";
	case OP_VECTORLIST:	return "OP_VECTORLIST";
	case OP_ASSIGN:		return "OP_ASSIGN";
	case OP_ADDASSIGN:	return "OP_ADDASSIGN";
	case OP_ADDORASSIGN:	return "OP_ADDORASSIGN";
	case OP_SUBASSIGN:	return "OP_SUBASSIGN";
	case OP_SUBORASSIGN:	return "OP_SUBORASSIGN";
	case OP_MULASSIGN:	return "OP_MULASSIGN";
	case OP_DIVASSIGN:	return "OP_DIVASSIGN";
	case OP_MODASSIGN:	return "OP_MODASSIGN";
	case OP_SHLASSIGN:	return "OP_SHLASSIGN";
	case OP_SHRASSIGN:	return "OP_SHRASSIGN";
	case OP_INC:		return "OP_INC";
	case OP_DEC:		return "OP_DEC";
	case OP_PREINC:		return "OP_PREINC";
	case OP_PREDEC:		return "OP_PREDEC";
	case OP_POSTINC:	return "OP_POSTINC";
	case OP_POSTDEC:	return "OP_POSTDEC";
	case OP_DEREF:		return "OP_DEREF";
	case OP_CONDEXPR:	return "OP_CONDEXPR";
	}
	return "<unknown expression operation>";
}

static const char *unitname(int n)
{
	switch(n) {
	case UNIT_NONE:	return "UNIT_NONE";
	case UNIT_MM:	return "UNIT_MM";
	case UNIT_IN:	return "UNIT_IN";
	case UNIT_DEG:	return "UNIT_DEG";
	case UNIT_RAD:	return "UNIT_RAD";
	}
	return "<unknown unit>";
}

static void unit_check(const node_t *n, const value_t *v)
{
	switch(v->unit) {
	case UNIT_NONE:
	case UNIT_MM:
	case UNIT_IN:
	case UNIT_DEG:
	case UNIT_RAD:
		break;
	default:
		rtinternal(n, "Invalid unit value %d", v->unit);
	}
}

static value_t *op_add_common(const node_t *ref, value_t *lv, const value_t *rv, const char *fn, opfunc_t opself)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Special case lv == VAL_STRING */
	if(isstring(lv)) {
		value_t *v = value_to_string(ref, rv);
		lv->str = realloc(lv->str, strlen(lv->str) + strlen(v->str) + 1);
		assert(lv->str != NULL);
		strcat(lv->str, v->str);
		value_delete(v);
		return lv;
	}

	/* Type check
	 * scalar + scalar
	 * vector + vector
	 * vectorlist + vector
	 * vectorlist + vectorlist	(append)
	 */
	if(!((isscalar(lv) && isscalar(rv)) || (isvector(lv) && isvector(rv)) || (isvectorlist(lv) && (isvector(rv) || isvectorlist(rv))))) {
		rterror(ref, "Scalars and vectors can only be added to scalars vectors and vector lists (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '+': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '+': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i += rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i += rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i + (double)rv->i * 25.4;
			else
				lv->d = (double)lv->i + (double)rv->i / 25.4;
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i + (double)rv->i * M_PI / 180.0;
			else
				lv->d = (double)lv->i + (double)rv->i / M_PI * 180.0;
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in %s %d, %d", fn, lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d += dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d += dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d += dval * 25.4;
			else
				lv->d += dval / 25.4;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d += dval * M_PI / 180.0;
			else
				lv->d += dval / M_PI * 180.0;
		} else
			rtinternal(ref, "Unit mismatch in %s %d, %d", fn, lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		while(lv->v.n < rv->v.n)
			value_vector_add_undef(lv);
		for(i = 0; i < lv->v.n && i < rv->v.n; i++)
			opself(ref, lv->v.vals[i], rv->v.vals[i]);
		break;
	case VAL_VECTORLIST:
		if(rv->type == VAL_VECTOR) {
			for(i = 0; i < lv->vl.n; i++)
				opself(ref, lv->vl.vecs[i], rv);
		} else {
			assert(rv->type == VAL_VECTORLIST);
			/* Append vectorlist, transfer from rv to lv */
			for(i = 0; i < rv->vl.n; i++)
				value_vectorlist_add(lv, value_dup(rv->vl.vecs[i]));
		}
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in %s %s", fn, valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_addor(const node_t *ref, value_t *lv, const value_t *rv)
{
	assert(lv != NULL);
	assert(rv != NULL);
	if(!isstring(lv) && rv->type == VAL_UNDEF) {
		/* x +| "undef" = x + 0 = x */
		return lv;
	}
	if(lv->type == VAL_UNDEF) {
		if(!isscalar(rv)) {
			rterror(ref, "Operation '<undef> + non-scalar' not supported");
			return lv;
		}
		/* "undef" +| x = 0 + x */
		*lv = *rv;
		return lv;
	}

	return op_add_common(ref, lv, rv, "op_addor", op_addor);
}

static value_t *op_add(const node_t *ref, value_t *lv, const value_t *rv)
{
	assert(lv != NULL);
	assert(rv != NULL);
	if(!isstring(lv) && lv->type == VAL_UNDEF) {
		/* "undef" + x = "undef" */
		return lv;
	}
	if(rv->type == VAL_UNDEF) {
		/* x + "undef" = x */
		return lv;
	}

	return op_add_common(ref, lv, rv, "op_add", op_add);
}

static value_t *op_sub_common(const node_t *ref, value_t *lv, const value_t *rv, const char *fn, opfunc_t opself)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Type check
	 * scalar - scalar
	 * vector - vector
	 * vectorlist - vector
	 */
	if(!((isscalar(lv) && isscalar(rv)) || ((isvector(lv) || isvectorlist(lv)) && isvector(rv)))) {
		rterror(ref, "Only scalars/vectors can be subtracted from scalars/vectors (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}
	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '-': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '-': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i -= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i -= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i - (double)rv->i * 25.4;
			else
				lv->d = (double)lv->i - (double)rv->i / 25.4;
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i - (double)rv->i * M_PI / 180.0;
			else
				lv->d = (double)lv->i - (double)rv->i / M_PI * 180.0;
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in %s %d, %d", fn, lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d -= dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d -= dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d -= dval * 25.4;
			else
				lv->d -= dval / 25.4;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d -= dval * M_PI / 180.0;
			else
				lv->d -= dval / M_PI * 180.0;
		} else
			rtinternal(ref, "Unit mismatch in %s %d, %d", fn, lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		while(lv->v.n < rv->v.n)
			value_vector_add_undef(lv);
		for(i = 0; i < lv->v.n && i < rv->v.n; i++)
			opself(ref, lv->v.vals[i], rv->v.vals[i]);
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < lv->vl.n; i++)
			opself(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in %s %s", fn, valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_subor(const node_t *ref, value_t *lv, const value_t *rv)
{
	assert(lv != NULL);
	assert(rv != NULL);
	if(rv->type == VAL_UNDEF) {
		/* x - "undef" = x - 0 = x */
		return lv;
	}
	if(lv->type == VAL_UNDEF) {
		/* "undef" - x = 0 - x */
		lv->type = VAL_INT;
		lv->i = 0;
		lv->unit = UNIT_NONE;
	}
	return op_sub_common(ref, lv, rv, "op_subor", op_subor);
}

static value_t *op_sub(const node_t *ref, value_t *lv, const value_t *rv)
{
	assert(lv != NULL);
	assert(rv != NULL);
	if(lv->type == VAL_UNDEF) {
		/* "undef" - x = "undef" */
		return lv;
	}
	if(rv->type == VAL_UNDEF) {
		/* x - "undef" = x */
		return lv;
	}
	return op_sub_common(ref, lv, rv, "op_sub", op_sub);
}

value_t *op_mul(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Multiply by "undef" = "undef" */
	if(lv->type == VAL_UNDEF)
		return lv;
	if(rv->type == VAL_UNDEF) {
		if(!isscalar(lv)) {
			rterror(ref, "Operation 'non-scalar * <undef>' not supported");
			return lv;
		}
		lv->type = VAL_UNDEF;
		return lv;
	}

	/* Type check
	 * scalar * scalar
	 * vector * scalar
	 * vector * vector	(dot product)
	 * vectorlist * scalar	(each vector multiplied by scalar)
	 */
	if(!(	(isscalar(lv) && isscalar(rv)) ||
		(isvector(lv) && (isscalar(rv) || isvector(rv))) ||
		(isvectorlist(lv) && isscalar(rv)))) {
		rterror(ref, "Scalars and vectors can only be multiplied by scalars and vectors (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '*': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '*': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i *= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i *= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i * (double)rv->i * 25.4;
			else
				lv->d = (double)lv->i * (double)rv->i / 25.4;
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i * (double)rv->i * M_PI / 180.0;
			else
				lv->d = (double)lv->i * (double)rv->i / M_PI * 180.0;
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in op_mul %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d *= dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d *= dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d *= dval * 25.4;
			else
				lv->d *= dval / 25.4;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d *= dval * M_PI / 180.0;
			else
				lv->d *= dval / M_PI * 180.0;
		} else
			rtinternal(ref, "Unit mismatch in op_mul %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		if(isscalar(rv)) {
			for(i = 0; i < lv->v.n; i++) {
				op_mul(ref, lv->v.vals[i], rv);
			}
		} else {
			/* Vector * vector == scalar (dot product) */
			double sum = 0.0;
			for(i = 0; i < lv->v.n && i < rv->v.n; i++) {
				sum += value_to_mm(lv->v.vals[i]) * value_to_mm(rv->v.vals[i]);
			}
			value_delete_vector(lv);
			lv->type = VAL_FLOAT;
			lv->unit = UNIT_NONE;
			lv->d = sum;
		}
		break;
	case VAL_VECTORLIST:
		assert(isscalar(rv));
		for(i = 0; i < lv->vl.n; i++)
			op_mul(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_mul %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_div(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Divide by "undef" = "undef" */
	if(lv->type == VAL_UNDEF)
		return lv;
	if(rv->type == VAL_UNDEF) {
		if(!isscalar(lv)) {
			rterror(ref, "Operation 'non-scalar / <undef>' not supported");
			return lv;
		}
		lv->type = VAL_UNDEF;
		return lv;
	}

	/* Type check
	 * scalar / scalar
	 * vector / scalar
	 * vectorlist / scalar
	 */
	if(!isscalar(rv) || !(isscalar(lv) || isvector(lv) || isvectorlist(lv))) {
		rterror(ref, "Scalars, vectors and vectorlists can only be divided by scalars (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '/': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '/': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(!rv->i)
			rterror(ref, "Divide by zero");
		if(lv->unit == rv->unit || unitmismatch) {
			lv->i /= rv->i;
			lv->unit = UNIT_NONE;
		} else if(rv->unit == UNIT_NONE) {
			lv->i /= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i /= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = (double)lv->i / ((double)rv->i * 25.4);
			else
				lv->d = (double)lv->i / ((double)rv->i / 25.4);
			lv->type = VAL_FLOAT;
			lv->unit = UNIT_NONE;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = (double)lv->i / ((double)rv->i * M_PI / 180.0);
			else
				lv->d = (double)lv->i / ((double)rv->i / M_PI * 180.0);
			lv->type = VAL_FLOAT;
			lv->unit = UNIT_NONE;
		} else
			rtinternal(ref, "Unit mismatch in op_div %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(fabs(dval) < EPSILON)
			rterror(ref, "Divide by zero (|val| < %e)", EPSILON);
		if(lv->unit == rv->unit || unitmismatch) {
			lv->d /= dval;
			lv->unit = UNIT_NONE;
		} else if(rv->unit == UNIT_NONE) {
			lv->d /= dval;
		} else if(lv->unit == UNIT_NONE) {
			lv->d /= dval;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d /= dval * 25.4;
			else
				lv->d /= dval / 25.4;
			lv->unit = UNIT_NONE;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d /= dval * M_PI / 180.0;
			else
				lv->d /= dval / M_PI * 180.0;
			lv->unit = UNIT_NONE;
		} else
			rtinternal(ref, "Unit mismatch in op_div %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		for(i = 0; i < lv->v.n; i++)
			op_div(ref, lv->v.vals[i], rv);
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < lv->vl.n; i++)
			op_div(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_div %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_mod(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	double dval;
	int unitmismatch = 0;
	assert(lv != NULL);
	assert(rv != NULL);

	/* Modulo "undef" = "undef" */
	if(lv->type == VAL_UNDEF)
		return lv;
	if(rv->type == VAL_UNDEF) {
		if(!isscalar(lv)) {
			rterror(ref, "Operation 'non-scalar %% <undef>' not supported");
			return lv;
		}
		lv->type = VAL_UNDEF;
		return lv;
	}

	/* Type check
	 * scalar % scalar
	 * vector % scalar
	 * vectorlist % scalar
	 */
	if(!isscalar(rv) || !(isscalar(lv) || isvector(lv) || isvectorlist(lv))) {
		rterror(ref, "Scalar modulo can only be calculated on scalars, vectors and vectorlists (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(isangle(lv) && isdistance(rv)) {
		rtwarning(ref, "Operation '%%': LHS has angular units and RHS has distance units, RHS units ignored");
		unitmismatch = 1;
	} else if(isdistance(lv) && isangle(rv)) {
		rtwarning(ref, "Operation '%%': LHS has distance units and RHS has angular units, RHS units ignored");
		unitmismatch = 1;
	}

	if(lv->type == VAL_INT && rv->type == VAL_FLOAT) {
		lv->d = lv->i;
		lv->type = VAL_FLOAT;
	}

	switch(lv->type) {
	case VAL_INT:
		assert(isscalar(rv));
		if(!rv->i)
			rterror(ref, "Modulo by zero");
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->i %= rv->i;
		} else if(lv->unit == UNIT_NONE) {
			lv->i %= rv->i;
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = fmod((double)lv->i, ((double)rv->i * 25.4));
			else
				lv->d = fmod((double)lv->i, ((double)rv->i / 25.4));
			lv->type = VAL_FLOAT;
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = fmod((double)lv->i, ((double)rv->i * M_PI / 180.0));
			else
				lv->d = fmod((double)lv->i, ((double)rv->i / M_PI * 180.0));
			lv->type = VAL_FLOAT;
		} else
			rtinternal(ref, "Unit mismatch in op_mod %d, %d", lv->unit, rv->unit);
		break;
	case VAL_FLOAT:
		assert(isscalar(rv));
		if(rv->type == VAL_INT)
			dval = rv->i;
		else
			dval = rv->d;
		if(fabs(dval) < EPSILON)
			rterror(ref, "Modulo by zero (|val| < %e)", EPSILON);
		if(lv->unit == rv->unit || rv->unit == UNIT_NONE || unitmismatch) {
			lv->d = fmod(lv->d, dval);
		} else if(lv->unit == UNIT_NONE) {
			lv->d = fmod(lv->d, dval);
			lv->unit = rv->unit;
		} else if(isdistance(lv)) {
			if(rv->unit == UNIT_IN)
				lv->d = fmod(lv->d, (dval * 25.4));
			else
				lv->d = fmod(lv->d, (dval / 25.4));
		} else if(isangle(lv)) {
			if(rv->unit == UNIT_DEG)
				lv->d = fmod(lv->d, (dval * M_PI / 180.0));
			else
				lv->d = fmod(lv->d, (dval / M_PI * 180.0));
		} else
			rtinternal(ref, "Unit mismatch in op_mod %d, %d", lv->unit, rv->unit);
		break;
	case VAL_VECTOR:
		for(i = 0; i < lv->v.n; i++)
			op_mod(ref, lv->v.vals[i], rv);
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < lv->vl.n; i++)
			op_mod(ref, lv->vl.vecs[i], rv);
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_mod %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_shl(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	int sft;
	assert(lv != NULL);
	assert(rv != NULL);

	if(lv->type == VAL_UNDEF || rv->type == VAL_UNDEF) {
		/* "undef" << x = undef */
		/* x << "undef" = x << 0 = x */
		return lv;
	}

	/* Type check
	 * scalar << scalar
	 * vector << scalar
	 * vectorlist << scalar
	 */
	if(!(isscalar(rv) && (isscalar(lv) || isvector(lv) || isvectorlist(lv)))) {
		rterror(ref, "Scalars, vectors and vectorlists can only be shifted by scalars (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(rv->unit != UNIT_NONE)
		rtwarning(ref, "Operation '<<': Shift count has units associated, ignored");
	if(rv->type == VAL_FLOAT && rv->d > (double)INT_MAX) {
		rterror(ref, "Operation '<<': Shift count floating point larger than integer can hold");
		return lv;
	}
	sft = value_to_int(rv);
	if(sft < 0) {
		rterror(ref, "Operation '<<': Shift count less than zero not allowed");
		return lv;
	}

	if(isscalar(lv) && sft > (int)(sizeof(int) * 8 - 1)) {
		rterror(ref, "Operation '<<': Shift count > %d has no effect", (int)(sizeof(int) * 8 - 1));
		return lv;
	} else if(!isscalar(lv) && sft > WARNSHIFTS)
		rtwarning(ref, "Operation '<<': Shift count > %d might be too large", sft);

	switch(lv->type) {
	case VAL_INT:
		lv->i <<= sft;
		break;
	case VAL_FLOAT:
		lv->d *= (double)(1 << sft);
		break;
	case VAL_VECTOR:
		for(i = 0; i < sft && lv->v.n; i++) {
			value_delete(lv->v.vals[0]);
			if(lv->v.n > 1)
				memmove(&lv->v.vals[0], &lv->v.vals[1], (lv->v.n-1)*sizeof(lv->v.vals[0]));
			lv->v.n--;
		}
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < sft && lv->vl.n; i++) {
			value_delete(lv->vl.vecs[0]);
			if(lv->vl.n > 1)
				memmove(&lv->vl.vecs[0], &lv->vl.vecs[1], (lv->vl.n-1)*sizeof(lv->vl.vecs[0]));
			lv->vl.n--;
		}
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_shl %s", valuetypename(lv->type));
	}
	return lv;
}

static value_t *op_shr(const node_t *ref, value_t *lv, const value_t *rv)
{
	int i;
	int sft;
	assert(lv != NULL);
	assert(rv != NULL);

	if(lv->type == VAL_UNDEF || rv->type == VAL_UNDEF) {
		/* "undef" >> x = undef */
		/* x >> "undef" = x << 0 = x */
		return lv;
	}

	/* Type check
	 * scalar >> scalar
	 * vector >> scalar
	 * vectorlist > scalar
	 */
	if(!(isscalar(rv) && (isscalar(lv) || isvector(lv) || isvectorlist(lv)))) {
		rterror(ref, "Scalars, vectors and vectorlists can only be shifted by scalars (lv=%s, rv=%s)", valuetypename(lv->type), valuetypename(rv->type));
		return lv;
	}

	unit_check(ref, lv);
	unit_check(ref, rv);

	if(rv->unit != UNIT_NONE)
		rtwarning(ref, "Operation '>>': Shift count has units associated, ignored");
	if(rv->type == VAL_FLOAT && rv->d > (double)INT_MAX) {
		rterror(ref, "Operation '>>': Shift count floating point larger than integer can hold");
		return lv;
	}
	sft = value_to_int(rv);
	if(sft < 0) {
		rterror(ref, "Operation '>>': Shift count less than zero not allowed");
		return lv;
	}

	if(isscalar(lv) && sft > (int)(sizeof(int) * 8 - 1)) {
		rterror(ref, "Operation '>>': Shift count > %d has no effect", (int)(sizeof(int) * 8 - 1));
		return lv;
	} else if(!isscalar(lv) && sft > WARNSHIFTS)
		rtwarning(ref, "Operation '>>': Shift count > %d might be too large", sft);

	switch(lv->type) {
	case VAL_INT:
		lv->i >>= sft;
		break;
	case VAL_FLOAT:
		lv->d /= (double)(1 << sft);
		break;
	case VAL_VECTOR:
		for(i = 0; i < sft; i++)
			value_vector_prepend(lv, value_new(VAL_UNDEF));
		break;
	case VAL_VECTORLIST:
		for(i = 0; i < sft; i++)
			value_vectorlist_prepend(lv, value_new(VAL_VECTOR));
		break;
	case VAL_STRING:
	default:
		rtinternal(ref, "Invalid value type in op_shl %s", valuetypename(lv->type));
	}
	return lv;
}

static int funcnamecmp(const void *a, const void *b)
{
	return strcmp(((const builtins_t *)a)->id, ((const builtins_t *)b)->id);
}

const builtins_t *find_builtin(const char *s)
{
	const builtins_t *bits = NULL;
	int nbits = 0;
	builtins_t key;
	const builtins_t *func;

	key.id = s;

	switch(cl_format) {
	case FMT_GCODE:
		bits = builtins_gcode;
		nbits = nbuiltins_gcode;
		break;
	case FMT_SVG:
		bits = builtins_svg;
		nbits = nbuiltins_svg;
		break;
	case FMT_DXF:
		bits = builtins_dxf;
		nbits = nbuiltins_dxf;
		break;
	default:
		rtinternal(NULL, "Undefined backend ID (%d) requested", cl_format);
	}
	/* Backend dependent functions */
	/* Searching these first allows for overloading */
	if((func = (const builtins_t *)bsearch(&key, bits, nbits, sizeof(*bits), funcnamecmp)))
		return func;

	/* Generic functions */
	if((func = (const builtins_t *)bsearch(&key, builtins_gcmc, nbuiltins_gcmc, sizeof(*builtins_gcmc), funcnamecmp)))
		return func;

	return NULL;
}

static int userfuncnamecmp(const void *a, const void *b)
{
	return strcmp((*(const node_t **)a)->func.id, (*(const node_t **)b)->func.id);
}

static const function_t *find_userfunc(const char *s)
{
	if(nfunctions < NSORTFUNC) {
		int i;
		for(i = 0; i < nfunctions; i++) {
			if(!strcmp(s, functions[i]->func.id))
				return &functions[i]->func;
		}
	} else {
		node_t n;
		node_t *np = &n;
		node_t **npp;
		n.func.id = (char *)s;
		npp = (node_t **)bsearch(&np, functions, nfunctions, sizeof(*functions), userfuncnamecmp);
		if(npp)
			return &(*npp)->func;
	}
	return NULL;
}

static int is_ident_expr(const node_t *n)
{
	if(n->type != NT_EXPR)
		return 0;
	return n->expr.op == OP_DEREF;
}

static value_t *op_call(node_t *n)
{
	value_t *v = NULL;
	int i, j;
	int nargs;
	const builtins_t *bifn;
	const function_t *ft;

	assert(n->call.args == NULL || n->call.args->type == NT_EXPRLIST);

	nargs = n->call.args ? n->call.args->nlist.n : 0;

	/* Push all arguments onto the value stack */
	for(i = 0; i < nargs; i++)
		interpret_expr(n->call.args->nlist.nodes[i]);

	/* Check for builtin function */
	if((bifn = find_builtin(n->call.id))) {
		v = bifn->func(n, nargs);
		if(!v)
			rtinternal(n, "Built-in function returned no value (NULL)");
		for(i = 0; i < nargs; i++)
			value_delete(value_pop(n));
		return v;
	}

	/* Script defined functions */
	if((ft = find_userfunc(n->call.id))) {
		assert(!ft->args || ft->args->type == NT_ARGLIST);
		if((ft->args && ft->args->alist.n != nargs) || (!ft->args && nargs))
			rterror(n, "Incorrect number of arguments in call to '%s'", n->call.id);
		assert(ft->args == NULL || nargs == ft->args->alist.n);
		/* Open new scope and push the arguments as local variables */
		variable_scope_push();
		for(j = 0; j < nargs; j++) {
			int revargc = ft->args->alist.n - j - 1;	/* Reverse order processing due to value stack */
			const arg_t *farg = &ft->args->alist.args[revargc];
			const node_t *narg = n->call.args->nlist.nodes[revargc];
			if(farg->isref) {
				/* Pass by reference */
				value_delete(value_pop(n));	/* Discard the value which got evaluated above */
				if(!is_ident_expr(narg))
					rterror(n, "%s argument %d: Cannot pass expression by reference", n->call.id, revargc+1);
				if(!variable_ref_local(farg->id, narg->expr.id))
					rterror(n, "%s argument %d: No variable by name %s to map to %s", n->call.id, revargc+1, narg->expr.id, farg->id);
			} else {
				/* Pass by value */
				if(!variable_set_local(farg->id, value_pop(n)))
					rtinternal(n, "Failed to add local variable '%s'", farg->id);
			}
		}
		v = execute(ft->body, NULL);
		variable_scope_pop();
		return v;
	}

	rterror(n, "Function '%s' not found", n->call.id);
	return NULL;
}

static value_t *op_indexref(const node_t *n, const variable_t **varref, int *iref)
{
	value_t *idx;
	value_t *val;
	const variable_t *var;
	int i;
	assert(varref != NULL);
	assert(iref != NULL);
	/* Indexed lvalue */
	idx = value_pop(n);	/* index */
	val = value_pop(n);	/* value to assign */
	if(!isscalar(idx)) {
		rterror(n, "Index on value must be scalar");
		value_delete(idx);
		value_delete(val);
		return NULL;
	}
	if(idx->unit != UNIT_NONE)
		rtwarning(n, "Index has units associated, ignored");
	i = value_to_int(idx);
	value_delete(idx);
	if(i < 0) {
		rterror(n, "Variable index on '%s' must be >= 0", n->expr.id);
		value_delete(val);
		return NULL;
	}
	if(!(var = variable_lookup_deref(n->expr.id, 0))) {
		rterror(n, "Index variable '%s' undefined", n->expr.id);
		value_delete(val);
		return NULL;
	}
	if(isvector(var->v)) {
		if(!isscalar(val)) {
			rterror(n, "Rvalue must be scalar in vector indexed assignment");
			value_delete(val);
			return NULL;
		}
		if(i > 9)
			rtwarning(n, "Variable index on vector '%s' is larger than 9 (%d)", n->expr.id, i);
		/* Extend vector to fit index */
		while(var->v->v.n < i+1)
			value_vector_add_undef(var->v);
		*varref = var;
		*iref = i;
		return val;
	} else if(isvectorlist(var->v)) {
		if(!n->expr.dblidx && !isvector(val)) {
			rterror(n, "Rvalue must be vector in vectorlist indexed assignment");
			value_delete(val);
			return NULL;
		}
		/* Extend vectorlist to fit index */
		while(var->v->vl.n < i+1)
			value_vectorlist_add(var->v, value_new(VAL_VECTOR));
		*varref = var;
		*iref = i;
		return val;
	} else {
		rterror(n, "Variable '%s' must be vector or vectorlist for index", n->expr.id);
	}
	value_delete(val);
	return NULL;
}

static value_t *op_dblindex(const node_t *n, const variable_t *var, int idx, int *iref)
{
	int i;
	value_t *lv;
	value_t *vec;

	if(!n->expr.dblidx)
		return NULL;
	assert(var != NULL);
	assert(idx >= 0);
	assert(var->v->type == VAL_VECTORLIST);
	assert(var->v->vl.n > idx);

	interpret_expr(n->expr.dblidx);
	lv = value_pop(n);
	if(!isscalar(lv)) {
		rterror(n->expr.dblidx, "Secondary index must be scalar");
		return NULL;
	}
	if(lv->unit != UNIT_NONE)
		rtwarning(n->expr.dblidx, "Secondary index has units associated, ignored");
	i = value_to_int(lv);
	value_delete(lv);
	if(i > 9)
		rtwarning(n->expr.dblidx, "Secondary variable index on vector '%s' is larger than 9 (%d)", n->expr.id, i);

	vec = var->v->vl.vecs[idx];
	/* Extend vector to fit index */
	while(vec->v.n < i+1)
		value_vector_add_undef(vec);
	if(iref)
		*iref = i;
	return vec->v.vals[i];
}

static void op_xassign(const node_t *n, opfunc_t func)
{
	int i;
	const variable_t *var;
	value_t *lv;
	value_t *rv;
	if(n->expr.right) {
		if((lv = op_indexref(n, &var, &i))) {
			if(n->expr.dblidx && (!isvectorlist(var->v) || !isscalar(lv))) {
				if(!isscalar(lv))
					rterror(n, "Double index assignment must have scalar rvalue");
				else
					rterror(n, "Double index only allowed on vectorlist");
			} else if(isvector(var->v))
				func(n, var->v->v.vals[i], lv);
			else if(isvectorlist(var->v)) {
				if((rv = op_dblindex(n, var, i, NULL)))
					func(n, rv, lv);
				else
					func(n, var->v->vl.vecs[i], lv);
			}
			value_delete(lv);
		}
	} else {
		if(!(var = variable_lookup_deref(n->expr.id, 0))) {
			rterror(n, "Assignment variable '%s' undefined", n->expr.id);
			value_delete(value_pop(n));
			return;
		}
		func(n, var->v, rv = value_pop(n));
		value_delete(rv);
	}
}

static void interpret_expr(node_t *n)
{
	int i, j;
	value_t *lv;
	value_t *rv;
	int lvb, rvb;
	double lvd, rvd;
	const variable_t *var;

	if(!n)
		return;

	if(n->type != NT_EXPR && n->type != NT_CALL)
		rtinternal(n, "Invalid node type %s (%d), expected NT_EXPR or NT_CALL", nodetypename(n->type), n->type);

	if(n->type == NT_CALL) {
		if(!(lv = op_call(n))) {
			rterror(n, "Function did not return a value");
			value_push(value_new(VAL_UNDEF));
		} else
			value_push(lv);
		return;
	}

	if(n->expr.op == OP_CONDEXPR) {
		/* Need the condition first, don't yet know which other to evaluate */
		interpret_expr(n->expr.cond);
	} else {
		interpret_expr(n->expr.left);
		/* Skip the right hand side initially on boolean ops to do short-circuit evaluation */
		if(n->expr.op != OP_LAND && n->expr.op != OP_LOR)
			interpret_expr(n->expr.right);
	}
	switch(n->expr.op) {
	case OP_ADD:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_add(n, lv, rv));
		value_delete(rv);
		break;
	case OP_ADDOR:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_addor(n, lv, rv));
		value_delete(rv);
		break;
	case OP_SUB:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_sub(n, lv, rv));
		value_delete(rv);
		break;
	case OP_SUBOR:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_subor(n, lv, rv));
		value_delete(rv);
		break;
	case OP_MUL:
		rv = value_pop(n);
		lv = value_pop(n);
		/* scalar * vector[list] --> vector[list] * scalar */
		if(isscalar(lv) && (isvector(rv) || isvectorlist(rv))) {
			value_push(op_mul(n, rv, lv));
			value_delete(lv);
		} else {
			value_push(op_mul(n, lv, rv));
			value_delete(rv);
		}
		break;
	case OP_SHL:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_shl(n, lv, rv));
		value_delete(rv);
		break;
	case OP_SHR:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_shr(n, lv, rv));
		value_delete(rv);
		break;
	case OP_DIV:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_div(n, lv, rv));
		value_delete(rv);
		break;
	case OP_MOD:
		rv = value_pop(n);
		lv = value_pop(n);
		value_push(op_mod(n, lv, rv));
		value_delete(rv);
		break;
	case OP_INDEX:
		/* Left is value to index, right is index */
		rv = value_pop(n);
		lv = value_pop(n);
		if(!isvector(lv) && !isvectorlist(lv))
			rterror(n, "Source for [] operator is not vector or vectorlist");
		if(!isscalar(rv))
			rterror(n, "Index of [] operator is not scalar");
		if(rv->unit != UNIT_NONE)
			rtwarning(n, "Index has a unit (%s), ignored", unitname(rv->unit));

		i = value_to_int(rv);
		if(isvector(lv)) {
			if(i < -lv->v.n || i >= lv->v.n) {
				if(lv->v.n)
					rtwarning(n, "Vector index out of range (%d outside [%d..%d])", i, -lv->v.n, lv->v.n-1);
				else
					rtwarning(n, "Vector index out of range (vector has no entries)");
				value_push(value_new(VAL_UNDEF));
			} else {
				if(i < 0)
					value_push(value_dup(lv->v.vals[lv->v.n+i]));
				else
					value_push(value_dup(lv->v.vals[i]));
			}
		} else { 
			if(i < -lv->vl.n || i >= lv->vl.n) {
				if(lv->vl.n)
					rtwarning(n, "Vectorlist index out of range (%d outside [%d..%d])", i, -lv->vl.n, lv->vl.n-1);
				else
					rtwarning(n, "Vectorlist index out of range (vectorlist has no entries)");
				value_push(value_new(VAL_VECTOR));
			} else {
				if(i < 0)
					value_push(value_dup(lv->vl.vecs[lv->vl.n+i]));
				else
					value_push(value_dup(lv->vl.vecs[i]));
			}
		}
		break;
	case OP_LOR:
	case OP_LAND:
		/* Only LHS got evaluated for logical ops */
		lv = value_pop(n);
		lvb = value_to_boolean(lv);
		value_delete(lv);
		/* Short-circuit the boolean operation */
		if(n->expr.op == OP_LAND && !lvb) {
			/* false && <anything> is always false */
			value_push(lv = value_new(VAL_INT));
			lv->i = 0;
			break;
		} else if(n->expr.op == OP_LOR && lvb) {
			/* true || <anything> is always true */
			value_push(lv = value_new(VAL_INT));
			lv->i = 1;
			break;
		}
		/* Undecided on LHS, evaluate RHS */
		interpret_expr(n->expr.right);
		rv = value_pop(n);
		rvb = value_to_boolean(rv);
		value_delete(rv);
		value_push(lv = value_new(VAL_INT));
		switch(n->expr.op) {
		case OP_LOR:	lv->i = lvb || rvb; break;
		case OP_LAND:	lv->i = lvb && rvb; break;
		default:	lv->i = 0; break;
		}
		break;
	case OP_EQ:
	case OP_NE:
	case OP_GT:
	case OP_LT:
	case OP_GE:
	case OP_LE:
		rv = value_pop(n);
		lv = value_pop(n);
		if((isdistance(lv) && isangle(rv)) || (isangle(lv) && isdistance(rv))) {
			rtwarning(n, "Units mismatch in comparison, converted to <none> (LHS=%s, RHS=%s)", unitname(lv->unit), unitname(rv->unit));
			rvd = value_to_double(rv);
			lvd = value_to_double(lv);
		} else {

			if((isnone(lv) && !isnone(rv)) || (!isnone(lv) && isnone(rv)))
				rtwarning(n, "Possible unit mismatch; comparing units %s with none", unitname(isnone(lv) ? rv->unit : lv->unit));
			if(isdistance(lv) || isdistance(rv)) {
				rvd = value_to_mm(rv);
				lvd = value_to_mm(lv);
			} else if(isangle(lv) || isangle(rv)) {
				rvd = value_to_deg(rv);
				lvd = value_to_deg(lv);
			} else {
				rvd = value_to_double(rv);
				lvd = value_to_double(lv);
			}
		}
		value_delete(rv);
		value_delete(lv);
		value_push(lv = value_new(VAL_INT));
		switch(n->expr.op) {
		case OP_EQ:	lv->i = fabs(lvd - rvd) < EPSILON; break;
		case OP_NE:	lv->i = fabs(lvd - rvd) >= EPSILON; break;
		case OP_GT:	lv->i = (lvd - rvd) > EPSILON; break;
		case OP_LT:	lv->i = (lvd - rvd) < -EPSILON; break;
		case OP_GE:	lv->i = (lvd - rvd) > -EPSILON; break;
		case OP_LE:	lv->i = (lvd - rvd) < EPSILON; break;
		default:	lv->i = 0; break;
		}
		break;
	case OP_NOT:
		lv = value_pop(n);
		lvb = value_to_boolean(lv);
		value_delete(lv);
		value_push(lv = value_new(VAL_INT));
		lv->i = !lvb;
		break;
	case OP_INT:
		value_push(value_new_int(n->expr.i, n->expr.unit));
		break;
	case OP_FLOAT:
		value_push(value_new_flt(n->expr.d, n->expr.unit));
		break;
	case OP_STRING:
		value_push(value_new_str(n->expr.id));
		break;
	case OP_VECTOR:
		lv = value_new(VAL_VECTOR);
		for(i = 0; i < n->expr.nlist.n; i++) {
			if(!n->expr.nlist.nodes[i])
				value_vector_add_undef(lv);
			else {
				interpret_expr(n->expr.nlist.nodes[i]);
				rv = value_pop(n->expr.nlist.nodes[i]);
				assert(rv != NULL);
				if(!isscalar(rv))
					rterror(n, "Vector entries must be scalar");
				value_vector_add(lv, rv);
			}
		}
		value_push(lv);
		break;
	case OP_VECTORLIST:
		lv = value_new(VAL_VECTORLIST);
		for(i = 0; i < n->expr.nlist.n; i++) {
			interpret_expr(n->expr.nlist.nodes[i]);
			rv = value_pop(n->expr.nlist.nodes[i]);
			assert(rv != NULL);
			if(rv->type != VAL_VECTOR)
				rterror(n, "Vectorlist entries must be vectors");
			value_vectorlist_add(lv, rv);
		}
		value_push(lv);
		break;
	case OP_ASSIGN:
		if(n->expr.right) {
			if((lv = op_indexref(n, &var, &i))) {
				if(n->expr.dblidx && (!isvectorlist(var->v) || !isscalar(lv))) {
					if(!isscalar(lv))
						rterror(n, "Double index assignment must have scalar rvalue");
					else
						rterror(n, "Double index only allowed on vectorlist");
				} else if(isvector(var->v)) {
					value_delete(var->v->v.vals[i]);
					var->v->v.vals[i] = lv;
				} else if(isvectorlist(var->v)) {
					if((rv = op_dblindex(n, var, i, &j))) {
						value_delete(var->v->vl.vecs[i]->v.vals[j]);
						var->v->vl.vecs[i]->v.vals[j] = lv;
					} else {
						value_delete(var->v->vl.vecs[i]);
						var->v->vl.vecs[i] = lv;
					}
				}
			}
		} else {
			variable_set(n->expr.id, value_pop(n));
		}
		break;
	case OP_ADDASSIGN:
		op_xassign(n, op_add);
		break;
	case OP_ADDORASSIGN:
		op_xassign(n, op_addor);
		break;
	case OP_SUBASSIGN:
		op_xassign(n, op_sub);
		break;
	case OP_SUBORASSIGN:
		op_xassign(n, op_subor);
		break;
	case OP_MULASSIGN:
		op_xassign(n, op_mul);
		break;
	case OP_DIVASSIGN:
		op_xassign(n, op_div);
		break;
	case OP_MODASSIGN:
		op_xassign(n, op_mod);
		break;
	case OP_SHLASSIGN:
		op_xassign(n, op_shl);
		break;
	case OP_SHRASSIGN:
		op_xassign(n, op_shr);
		break;
	case OP_INC:
	case OP_PREINC:
	case OP_POSTINC:
		if(!(var = variable_lookup_deref(n->expr.id, 0))) {
			rterror(n, "Increment variable '%s' undefined", n->expr.id);
			if(n->expr.op != OP_INC)
				value_push(value_new(VAL_UNDEF));
			break;
		}
		if(n->expr.op == OP_POSTINC)
			value_push(value_dup(var->v));
		switch(var->v->type) {
		case VAL_INT:	var->v->i++; break;
		case VAL_FLOAT:	var->v->d += 1.0; break;
		default:
			rtwarning(n, "Increment on non-scalar has no effect");
			break;
		}
		if(n->expr.op == OP_PREINC)
			value_push(value_dup(var->v));
		break;
	case OP_DEC:
	case OP_PREDEC:
	case OP_POSTDEC:
		if(!(var = variable_lookup_deref(n->expr.id, 0))) {
			rterror(n, "Decrement variable '%s' undefined", n->expr.id);
			if(n->expr.op != OP_DEC)
				value_push(value_new(VAL_UNDEF));
			break;
		}
		if(n->expr.op == OP_POSTDEC)
			value_push(value_dup(var->v));
		switch(var->v->type) {
		case VAL_INT:	var->v->i--; break;
		case VAL_FLOAT:	var->v->d -= 1.0; break;
		default:
			rtwarning(n, "Increment on non-scalar has no effect");
			break;
		}
		if(n->expr.op == OP_PREDEC)
			value_push(value_dup(var->v));
		break;
	case OP_DEREF:
		if(!(var = variable_lookup_deref(n->expr.id, 0))) {
			rterror(n, "Variable '%s' undefined", n->expr.id);
			value_push(value_new(VAL_UNDEF));
			break;
		}
		value_push(value_dup(var->v));
		break;
	case OP_CONDEXPR:	/* ?: */
		lv = value_pop(n);
		lvb = value_to_boolean(lv);
		value_delete(lv);
		if(lvb)
			interpret_expr(n->expr.left);	/* true clause */
		else
			interpret_expr(n->expr.right);	/* false clause */
		break;
	default:
		rtinternal(n, "Invalid expression operation %s (%d)", expropname(n->expr.op), n->expr.op);
		break;
	}
}

static value_t *op_foreach(node_t *n)
{
	int i;
	value_t *v = NULL;
	value_t *src;
	node_t *brk = NULL;
	assert(n->lfe.src->type == NT_EXPR || n->lfe.src->type == NT_CALL);
	interpret_expr(n->lfe.src);
	src = value_pop(n);
	if(!isvectorlist(src))
		rterror(n, "Foreach's source must be vectorlist");
	for(i = 0; i < src->vl.n; i++) {
		value_t *val = value_dup(src->vl.vecs[i]);
		assert(val->type == VAL_VECTOR);
		variable_set(n->lfe.dst, val);
		v = execute(n->lfe.stmts, &brk);
		if(v)
			break;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
	}
	value_delete(src);
	return v;
}

static value_t *op_repeat(node_t *n)
{
	int repeats;
	int step;
	int i;
	value_t *v = NULL;
	value_t *src;
	node_t *brk = NULL;
	assert(n->lfe.src->type == NT_EXPR || n->lfe.src->type == NT_CALL);
	interpret_expr(n->lfe.src);
	src = value_pop(n);
	if(!isscalar(src))
		rterror(n, "Repeat's source must be scalar");
	if(!isnone(src))
		rtwarning(n, "Repeat's source has units associated, ignored");
	if(src->type == VAL_FLOAT && fabs(src->d - round(src->d)) > EPSILON)
		rtwarning(n, "Repeat's source is floating point and has a non-integer fraction, truncated to integer");
	repeats = value_to_int(src);

	if(repeats < 0) {
		repeats = -repeats;
		step = -1;
	} else
		step = 1;

	for(i = 0; i < repeats; i++) {
		if(n->lfe.dst) {
			value_t *val = value_new_int((i+1) * step, UNIT_NONE);
			variable_set(n->lfe.dst, val);
		}
		v = execute(n->lfe.stmts, &brk);
		if(v)
			break;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
	}
	value_delete(src);
	return v;
}

static int func_sort(const void *a, const void *b)
{
	return strcmp((*(const node_t **)a)->func.id, (*(const node_t **)b)->func.id);
}

static void function_add(node_t *n)
{
	int i;
	assert(n->type == NT_FUNCTION);
	testalloc((void **)&functions, nfunctions, &nafunctions, sizeof(functions[0]));
	for(i = 0; i < nfunctions; i++) {
		if(!strcmp(functions[i]->func.id, n->func.id)) {
			rterror(n, "Function '%s' already defined", n->func.id);
			return;
		}
	}
	if(find_builtin(n->func.id)) {
		rterror(n, "Function '%s' is a built-in function", n->func.id);
		return;
	}
	functions[nfunctions] = n;
	nfunctions++;
	qsort(functions, nfunctions, sizeof(*functions), func_sort);
}

static value_t *op_for(node_t *n)
{
	node_t *brk = NULL;
	value_t *v;
	int b;
	assert(n->lfor.cond->type == NT_EXPR || n->lfor.cond->type == NT_CALL);

	if(n->lfor.init) {
		if(execute(n->lfor.init, NULL))
			rterror(n, "For-loop initialization should not return a value");
	}
	while(1) {
		interpret_expr(n->lfor.cond);
		v = value_pop(n);
		b = value_to_boolean(v);
		value_delete(v);
		if(!b)
			break;
		if((v = execute(n->lfor.stmts, &brk)))
			return v;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
		if(n->lfor.inc) {
			if(execute(n->lfor.inc, NULL))
				rterror(n, "For-loop increment should not return a value");
		}
	}
	return NULL;
}

static value_t *op_while(node_t *n)
{
	node_t *brk = NULL;
	value_t *v;
	int b;
	assert(n->lfor.cond->type == NT_EXPR || n->lfor.cond->type == NT_CALL);

	while(1) {
		interpret_expr(n->lfor.cond);
		v = value_pop(n);
		b = value_to_boolean(v);
		value_delete(v);
		if(!b)
			break;
		if((v = execute(n->lfor.stmts, &brk)))
			return v;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
	}
	return NULL;
}

static value_t *op_dowhile(node_t *n)
{
	node_t *brk = NULL;
	value_t *v;
	int b;
	assert(n->lfor.cond->type == NT_EXPR || n->lfor.cond->type == NT_CALL);

	while(1) {
		if((v = execute(n->lfor.stmts, &brk)))
			return v;
		if(brk && brk->type == NT_BREAK)
			break;
		if(brk && brk->type == NT_CONTINUE)
			brk = NULL;
		else {
			interpret_expr(n->lfor.cond);
			v = value_pop(n);
			b = value_to_boolean(v);
			value_delete(v);
			if(!b)
				break;
		}
	}
	return NULL;
}

static value_t *op_if(node_t *n, node_t **brkref)
{
	value_t *v;
	int b;
	assert(n->cond.cond->type == NT_EXPR || n->cond.cond->type == NT_CALL);

	interpret_expr(n->cond.cond);
	v = value_pop(n);
	b = value_to_boolean(v);
	value_delete(v);
	if(b) {
		if((v = execute(n->cond.ifclause, brkref)))
			return v;
	} else if(n->cond.elifclauses) {
		node_t *e;
		for(e = n->cond.elifclauses; e; e = e->next) {
			assert(e->type == NT_ELIF);
			interpret_expr(e->cond.cond);
			v = value_pop(n);
			b = value_to_boolean(v);
			value_delete(v);
			if(b) {
				if((v = execute(e->cond.ifclause, brkref)))
					return v;
				break;
			}
		}
		/* Execute any trailing else-clause if no elif matches */
		if(!b && n->cond.elseclause) {
			if((v = execute(n->cond.elseclause, brkref)))
				return v;
		}
	} else if(n->cond.elseclause) {
		if((v = execute(n->cond.elseclause, brkref)))
			return v;
	}
	return NULL;
}

static value_t *interpret(node_t *n, node_t **brkref)
{
	value_t *v;
	int i;
	if(!n)
		return NULL;

	switch(n->type) {
	case NT_EXPRLIST:	/* Only as call argument */
	case NT_ARGLIST:	/* Only as function arg definition */
	case NT_LOCALLIST:	/* Only as local arg definition */
	case NT_BREAK:		/* Only in loops */
	case NT_CONTINUE:	/* Only in loops */
	default:
		rtinternal(n, "Invalid node type %s (%d)", nodetypename(n->type), n->type);
		break;
	case NT_EXPR:		interpret_expr(n); break;
	case NT_IF:		return op_if(n, brkref);
	case NT_FOR:		return op_for(n);
	case NT_FOREACH:	return op_foreach(n);
	case NT_REPEAT:		return op_repeat(n);
	case NT_WHILE:		return op_while(n);
	case NT_DOWHILE:	return op_dowhile(n);
	case NT_FUNCTION:	function_add(n); break;
	case NT_CALL:
		if((v = op_call(n)))
			value_delete(v);
		break;
	case NT_RETURN:
		if(n->eref) {
			interpret_expr(n->eref);
			return value_pop(n);
		} else
			return value_new(VAL_UNDEF);
	case NT_LOCAL:
		assert(n->eref != NULL && n->eref->type == NT_LOCALLIST);
		for(i = 0; i < n->eref->llist.n; i++) {
			value_t *v;
			if(n->eref->llist.locs[i].init) {
				interpret_expr(n->eref->llist.locs[i].init);
				v = value_pop(n);
			} else
				v = value_new(VAL_UNDEF);
			variable_set_local(n->eref->llist.locs[i].id, v);
		}
		break;
	}
	return NULL;
}

value_t *execute(node_t *head, node_t **brkref)
{
	node_t *n;
	value_t *v;
	for(n = head; n; n = n->next) {
		if(n->type == NT_BREAK || n->type == NT_CONTINUE) {
			if(brkref) {
				*brkref = n;
				return NULL;
			} else
				rterror(n, "Unexpected break or continue");
		}
		if((v = interpret(n, brkref)))
			return v;
		if(brkref && *brkref)
			return NULL;
	}
	return NULL;
}

